@@ -173,24 +173,24 @@
 
 #include "dispatch_common.h"
 
-#ifdef __APPLE__
+#if defined(__APPLE__)
 #define GLX_LIB "/opt/X11/lib/libGL.1.dylib"
+#define OPENGL_LIB "/System/Library/Frameworks/OpenGL.framework/Versions/Current/OpenGL"
+#define GLES1_LIB "libGLESv1_CM.so"
+#define GLES2_LIB "libGLESv2.so"
 #elif defined(__ANDROID__)
 #define GLX_LIB "libGLESv2.so"
-#else
-#define GLVND_GLX_LIB "libGLX.so.1"
-#define GLX_LIB "libGL.so.1"
-#endif
-
-#ifdef __ANDROID__
 #define EGL_LIB "libEGL.so"
 #define GLES1_LIB "libGLESv1_CM.so"
 #define GLES2_LIB "libGLESv2.so"
-#elif defined _WIN32
+#elif defined(_WIN32)
 #define EGL_LIB "libEGL.dll"
 #define GLES1_LIB "libGLES_CM.dll"
 #define GLES2_LIB "libGLESv2.dll"
+#define OPENGL_LIB "OPENGL32"
 #else
+#define GLVND_GLX_LIB "libGLX.so.1"
+#define GLX_LIB "libGL.so.1"
 #define EGL_LIB "libEGL.so.1"
 #define GLES1_LIB "libGLESv1_CM.so.1"
 #define GLES2_LIB "libGLESv2.so.2"
@@ -290,7 +290,7 @@ library_init(void)
 }
 
 static bool
-get_dlopen_handle(void **handle, const char *lib_name, bool exit_on_fail)
+get_dlopen_handle(void **handle, const char *lib_name, bool exit_on_fail, bool load)
 {
     if (*handle)
         return true;
@@ -305,11 +305,15 @@ get_dlopen_handle(void **handle, const char *lib_name, bool exit_on_fail)
 #else
     pthread_mutex_lock(&api.mutex);
     if (!*handle) {
-        *handle = dlopen(lib_name, RTLD_LAZY | RTLD_LOCAL);
+        int flags = RTLD_LAZY | RTLD_LOCAL;
+        if (!load)
+            flags |= RTLD_NOLOAD;
+
+        *handle = dlopen(lib_name, flags);
         if (!*handle) {
             if (exit_on_fail) {
                 fprintf(stderr, "Couldn't open %s: %s\n", lib_name, dlerror());
-                exit(1);
+                abort();
             } else {
                 (void)dlerror();
             }
@@ -322,15 +326,11 @@ get_dlopen_handle(void **handle, const char *lib_name, bool exit_on_fail)
 }
 
 static void *
-do_dlsym(void **handle, const char *lib_name, const char *name,
-         bool exit_on_fail)
+do_dlsym(void **handle, const char *name, bool exit_on_fail)
 {
     void *result;
     const char *error = "";
 
-    if (!get_dlopen_handle(handle, lib_name, exit_on_fail))
-        return NULL;
-
 #ifdef _WIN32
     result = GetProcAddress(*handle, name);
 #else
@@ -339,8 +339,8 @@ do_dlsym(void **handle, const char *lib_name, const char *name,
         error = dlerror();
 #endif
     if (!result && exit_on_fail) {
-        fprintf(stderr,"%s() not found in %s: %s\n", name, lib_name, error);
-        exit(1);
+        fprintf(stderr, "%s() not found: %s\n", name, error);
+        abort();
     }
 
     return result;
@@ -392,10 +392,10 @@ epoxy_is_desktop_gl(void)
 }
 
 static int
-epoxy_internal_gl_version(GLenum version_string, int error_version)
+epoxy_internal_gl_version(GLenum version_string, int error_version, int factor)
 {
     const char *version = (const char *)glGetString(version_string);
-    GLint major, minor, factor;
+    GLint major, minor;
     int scanf_count;
 
     if (!version)
@@ -410,14 +410,9 @@ epoxy_internal_gl_version(GLenum version_string, int error_version)
     if (scanf_count != 2) {
         fprintf(stderr, "Unable to interpret GL_VERSION string: %s\n",
                 version);
-        exit(1);
+        abort();
     }
 
-    if (minor >= 10)
-        factor = 100;
-    else
-        factor = 10;
-
     return factor * major + minor;
 }
 
@@ -439,7 +434,7 @@ epoxy_internal_gl_version(GLenum version_string, int error_version)
 int
 epoxy_gl_version(void)
 {
-    return epoxy_internal_gl_version(GL_VERSION, 0);
+    return epoxy_internal_gl_version(GL_VERSION, 0, 10);
 }
 
 int
@@ -448,7 +443,7 @@ epoxy_conservative_gl_version(void)
     if (api.begin_count)
         return 100;
 
-    return epoxy_internal_gl_version(GL_VERSION, 100);
+    return epoxy_internal_gl_version(GL_VERSION, 100, 10);
 }
 
 /**
@@ -471,7 +466,7 @@ epoxy_glsl_version(void)
 {
     if (epoxy_gl_version() >= 20 ||
         epoxy_has_gl_extension ("GL_ARB_shading_language_100"))
-        return epoxy_internal_gl_version(GL_SHADING_LANGUAGE_VERSION, 0);
+        return epoxy_internal_gl_version(GL_SHADING_LANGUAGE_VERSION, 0, 100);
 
     return 0;
 }
@@ -548,16 +543,29 @@ epoxy_internal_has_gl_extension(const char *ext, bool invalid_op_mode)
     }
 }
 
+bool
+epoxy_load_glx(bool exit_if_fails, bool load)
+{
+#if PLATFORM_HAS_GLX
+# ifdef GLVND_GLX_LIB
+    /* prefer the glvnd library if it exists */
+    if (!api.glx_handle)
+	get_dlopen_handle(&api.glx_handle, GLVND_GLX_LIB, false, load);
+# endif
+    if (!api.glx_handle)
+        get_dlopen_handle(&api.glx_handle, GLX_LIB, exit_if_fails, load);
+#endif
+    return api.glx_handle != NULL;
+}
+
 void *
 epoxy_conservative_glx_dlsym(const char *name, bool exit_if_fails)
 {
-#ifdef GLVND_GLX_LIB
-    /* prefer the glvnd library if it exists */
-    if (!api.glx_handle)
-	get_dlopen_handle(&api.glx_handle, GLVND_GLX_LIB, false);
+#if PLATFORM_HAS_GLX
+    if (epoxy_load_glx(exit_if_fails, exit_if_fails))
+        return do_dlsym(&api.glx_handle, name, exit_if_fails);
 #endif
-
-    return do_dlsym(&api.glx_handle, GLX_LIB, name, exit_if_fails);
+    return NULL;
 }
 
 /**
@@ -572,23 +580,7 @@ epoxy_current_context_is_glx(void)
 #else
     void *sym;
 
-    /* If we've been called already, don't load more */
-    if (!api.egl_handle != !api.glx_handle) {
-	if (api.glx_handle)
-	    return true;
-	else if (api.egl_handle)
-	    return false;
-    }
-
-    /* If the application hasn't explicitly called some of our GLX
-     * or EGL code but has presumably set up a context on its own,
-     * then we need to figure out how to getprocaddress anyway.
-     *
-     * If there's a public GetProcAddress loaded in the
-     * application's namespace, then use that.
-     */
-
-    sym = dlsym(NULL, "glXGetCurrentContext");
+    sym = epoxy_conservative_glx_dlsym("glXGetCurrentContext", false);
     if (sym) {
         if (glXGetCurrentContext())
             return true;
@@ -597,7 +589,7 @@ epoxy_current_context_is_glx(void)
     }
 
 #if PLATFORM_HAS_EGL
-    sym = dlsym(NULL, "eglGetCurrentContext");
+    sym = epoxy_conservative_egl_dlsym("eglGetCurrentContext", false);
     if (sym) {
         if (epoxy_egl_get_current_gl_context_api() != EGL_NONE)
             return false;
@@ -606,21 +598,6 @@ epoxy_current_context_is_glx(void)
     }
 #endif /* PLATFORM_HAS_EGL */
 
-    /* OK, couldn't find anything in the app's address space.
-     * Presumably they dlopened with RTLD_LOCAL, which hides it
-     * from us.  Just go dlopen()ing likely libraries and try them.
-     */
-    sym = epoxy_conservative_glx_dlsym("glXGetCurrentContext", false);
-    if (sym && glXGetCurrentContext())
-        return true;
-
-#if PLATFORM_HAS_EGL
-    sym = do_dlsym(&api.egl_handle, EGL_LIB, "eglGetCurrentContext",
-                   false);
-    if (sym && epoxy_egl_get_current_gl_context_api() != EGL_NONE)
-        return false;
-#endif /* PLATFORM_HAS_EGL */
-
     return false;
 #endif /* PLATFORM_HAS_GLX */
 }
@@ -651,10 +628,24 @@ epoxy_conservative_has_gl_extension(const char *ext)
     return epoxy_internal_has_gl_extension(ext, true);
 }
 
+bool
+epoxy_load_egl(bool exit_if_fails, bool load)
+{
+#if PLATFORM_HAS_EGL
+    return get_dlopen_handle(&api.egl_handle, EGL_LIB, exit_if_fails, load);
+#else
+    return false;
+#endif
+}
+
 void *
 epoxy_conservative_egl_dlsym(const char *name, bool exit_if_fails)
 {
-    return do_dlsym(&api.egl_handle, EGL_LIB, name, exit_if_fails);
+#if PLATFORM_HAS_EGL
+    if (epoxy_load_egl(exit_if_fails, exit_if_fails))
+        return do_dlsym(&api.egl_handle, name, exit_if_fails);
+#endif
+    return NULL;
 }
 
 void *
@@ -669,31 +660,32 @@ epoxy_glx_dlsym(const char *name)
     return epoxy_conservative_glx_dlsym(name, true);
 }
 
+static void
+epoxy_load_gl(void)
+{
+    if (api.gl_handle)
+	return;
+
+#if defined(_WIN32) || defined(__APPLE__)
+    get_dlopen_handle(&api.gl_handle, OPENGL_LIB, true, true);
+#else
+
+#if defined(OPENGL_LIB)
+    if (!api.gl_handle)
+	get_dlopen_handle(&api.gl_handle, OPENGL_LIB, false, true);
+#endif
+
+    get_dlopen_handle(&api.glx_handle, GLX_LIB, true, true);
+    api.gl_handle = api.glx_handle;
+#endif
+}
+
 void *
 epoxy_gl_dlsym(const char *name)
 {
-#ifdef _WIN32
-    return do_dlsym(&api.gl_handle, "OPENGL32", name, true);
-#elif defined(__APPLE__)
-    return do_dlsym(&api.gl_handle,
-                    "/System/Library/Frameworks/OpenGL.framework/Versions/Current/OpenGL",
-                    name, true);
-#else
-    void *sym;
+    epoxy_load_gl();
 
-# if defined(OPENGL_LIB)
-    if (!api.gl_handle)
-	get_dlopen_handle(&api.gl_handle, OPENGL_LIB, false);
-# endif
-
-    if (api.gl_handle)
-	return do_dlsym(&api.gl_handle, NULL, name, true);
-
-    sym = do_dlsym(&api.glx_handle, GLX_LIB, name, true);
-    api.gl_handle = api.glx_handle; /* skip the dlopen next time */
-
-    return sym;
-#endif
+    return do_dlsym(&api.gl_handle, name, true);
 }
 
 void *
@@ -702,7 +694,8 @@ epoxy_gles1_dlsym(const char *name)
     if (epoxy_current_context_is_glx()) {
         return epoxy_get_proc_address(name);
     } else {
-        return do_dlsym(&api.gles1_handle, GLES1_LIB, name, true);
+        get_dlopen_handle(&api.gles1_handle, GLES1_LIB, true, true);
+        return do_dlsym(&api.gles1_handle, name, true);
     }
 }
 
@@ -712,7 +705,8 @@ epoxy_gles2_dlsym(const char *name)
     if (epoxy_current_context_is_glx()) {
         return epoxy_get_proc_address(name);
     } else {
-        return do_dlsym(&api.gles2_handle, GLES2_LIB, name, true);
+        get_dlopen_handle(&api.gles2_handle, GLES2_LIB, true, true);
+        return do_dlsym(&api.gles2_handle, name, true);
     }
 }
 
@@ -732,10 +726,12 @@ epoxy_gles3_dlsym(const char *name)
     if (epoxy_current_context_is_glx()) {
         return epoxy_get_proc_address(name);
     } else {
-        void *func = do_dlsym(&api.gles2_handle, GLES2_LIB, name, false);
+        if (get_dlopen_handle(&api.gles2_handle, GLES2_LIB, false, true)) {
+            void *func = do_dlsym(&api.gles2_handle, name, false);
 
-        if (func)
-            return func;
+            if (func)
+                return func;
+        }
 
         return epoxy_get_proc_address(name);
     }
@@ -810,22 +806,22 @@ epoxy_get_bootstrap_proc_address(const char *name)
      * non-X11 ES2 context from loading a bunch of X11 junk).
      */
 #if PLATFORM_HAS_EGL
-    get_dlopen_handle(&api.egl_handle, EGL_LIB, false);
+    get_dlopen_handle(&api.egl_handle, EGL_LIB, false, true);
     if (api.egl_handle) {
+        int version = 0;
         switch (epoxy_egl_get_current_gl_context_api()) {
         case EGL_OPENGL_API:
             return epoxy_gl_dlsym(name);
         case EGL_OPENGL_ES_API:
-            /* We can't resolve the GL version, because
-             * epoxy_glGetString() is one of the two things calling
-             * us.  Try the GLES2 implementation first, and fall back
-             * to GLES1 otherwise.
-             */
-            get_dlopen_handle(&api.gles2_handle, GLES2_LIB, false);
-            if (api.gles2_handle)
-                return epoxy_gles2_dlsym(name);
-            else
-                return epoxy_gles1_dlsym(name);
+            if (eglQueryContext(eglGetCurrentDisplay(),
+                                eglGetCurrentContext(),
+                                EGL_CONTEXT_CLIENT_VERSION,
+                                &version)) {
+                if (version >= 2)
+                    return epoxy_gles2_dlsym(name);
+                else
+                    return epoxy_gles1_dlsym(name);
+            }
         }
     }
 #endif /* PLATFORM_HAS_EGL */
